library gobang;

{ **********************************************
  This DLL contains a number of functions to play
  the game of gobang. In particular it contains
  the player AI. It is used as an extension package
  for a Gobang game in Game Maker.

  written by Mark Overmars
  last changed November 2, 2007
  ********************************************** }

uses
  SysUtils,
  Classes;

{$R *.res}

const FSIZE = 19;              // Field size
      MAXMOVE = FSIZE*FSIZE;   // Maximal number of possible moves
      POS_EMPTY = 0;
      POS_WHITE = 10;
      POS_BLACK = 100;
      POS_BORDER = 1000;

var field:array[-4..FSIZE+5,-4..FSIZE+5] of integer;      // field of positions
    winx,winy:array[1..5] of integer;                     // positions of winning moves
    xmoves,ymoves:array[1..FSIZE*FSIZE] of integer;       // all the moved made
    movenumb:integer;                                     // number of moves so far

{ ***********************************
  Some help functions
  *********************************** }

function empty(x,y:integer):boolean;
// Returns whether the position is empty
begin
  Result := field[x,y] = POS_EMPTY;
end;

function hor(x,y,n:integer):integer;
// Returns the sum of a horizonal line starting at (x,y) of length n
var i:integer;
begin
  Result := 0;
  for i:=0 to n-1 do Result := Result + field[x+i,y];
end;

function vert(x,y,n:integer):integer;
// Returns the sum of a vertical line starting at (x,y) of length n
var i:integer;
begin
  Result := 0;
  for i:=0 to n-1 do Result := Result + field[x,y+i];
end;

function up(x,y,n:integer):integer;
// Returns the sum of a diagonal up line starting at (x,y) of length n
var i:integer;
begin
  Result := 0;
  for i:=0 to n-1 do Result := Result + field[x+i,y+i];
end;

function down(x,y,n:integer):integer;
// Returns the sum of a diagonal down line starting at (x,y) of length n
var i:integer;
begin
  Result := 0;
  for i:=0 to n-1 do Result := Result + field[x+i,y-i];
end;


{ ***********************************
  Some basic tests
  *********************************** }

function CheckWin(x,y:integer):boolean;
// Returns whether the stone at position x,y is part of a winning row
// Fills in the the winx and winy arrays if so
var i,j,stone:integer;
begin
  Result := false;
  stone := field[x,y];
  for i:=-4 to 0 do
  begin
    if hor(x+i,y,5) = 5*stone then
    begin
      for j:=1 to 5 do begin winx[j] := x+i+j-1; winy[j] := y; end;
      Result := true;
    end;
    if vert(x,y+i,5) = 5*stone then
    begin
      for j:=1 to 5 do begin winx[j] := x; winy[j] := y+i+j-1; end;
      Result := true;
    end;
    if up(x+i,y+i,5) = 5*stone then
    begin
      for j:=1 to 5 do begin winx[j] := x+i+j-1; winy[j] := y+i+j-1; end;
      Result := true;
    end;
    if down(x+i,y-i,5) = 5*stone then
    begin
      for j:=1 to 5 do begin winx[j] := x+i+j-1; winy[j] := y-i-j+1; end;
      Result := true;
    end;
  end;
end;

function CheckKin5(x,y,stone,k:integer):integer;
// Returns in how many rows of length 5 including (x,y) there are k equal stones
var i,a,b,c:integer;
begin
  Result := 0;
  for i:=-4 to 0 do
  begin
    a := x+i; b:=y+i; c := y-i;
    if hor (a,y,5) = k*stone then Result := Result +1;
    if vert(x,b,5) = k*stone then Result := Result +1;
    if up  (a,b,5) = k*stone then Result := Result +1;
    if down(a,c,5) = k*stone then Result := Result +1;
  end;
end;

function CheckKin4(x,y,stone,k:integer):integer;
// Returns in how many rows of length 6 including (x,y) there is empty,k in 4,empty
var i,a,b,c,n1,n2,n3,n4:integer;
begin
  n1:=0; n2:=0; n3:=0; n4:=0;
  for i:=-3 to 0 do
  begin
    a := x+i; b:=y+i; c := y-i;
    if (hor (a,y,4) = k*stone) and empty(a-1,y)   and empty(a+4,y)   then n1 := 1;
    if (vert(x,b,4) = k*stone) and empty(x,b-1)   and empty(x,b+4)   then n2 := 1;
    if (up  (a,b,4) = k*stone) and empty(a-1,b-1) and empty(a+4,b+4) then n3 := 1;
    if (down(a,c,4) = k*stone) and empty(a-1,c+1) and empty(a+4,c-4) then n4 := 1;
  end;
  Result := n1+n2+n3+n4;
end;

function CheckKin3(x,y,stone,k:integer):integer;
// Returns in how many rows of length 5 including (x,y) there is empty,k in 3,empty
var i,a,b,c,n1,n2,n3,n4:integer;
begin
  n1:=0; n2:=0; n3:=0; n4:=0;
  for i:=-2 to 0 do
  begin
    a := x+i; b:=y+i; c := y-i;
    if (hor (a,y,3) = k*stone) and empty(a-1,y)   and empty(a+3,y)   then n1 := 1;
    if (vert(x,b,3) = k*stone) and empty(x,b-1)   and empty(x,b+3)   then n2 := 1;
    if (up  (a,b,3) = k*stone) and empty(a-1,b-1) and empty(a+3,b+3) then n3 := 1;
    if (down(a,c,3) = k*stone) and empty(a-1,c+1) and empty(a+3,c-3) then n4 := 1;
  end;
  Result := n1+n2+n3+n4;
end;

function CheckThrees(x,y,stone:integer):integer;
// Returns in how many open rows of length 3 (x,y) is involved
var i,a,b,c,n1,n2,n3,n4:integer;
begin
  n1:=0; n2:=0; n3:=0; n4:=0;
  for i:=-4 to 0 do
  begin
    a := x+i; b:=y+i; c := y-i;
    if (hor (a,y,5) = 4*stone) or((hor (a,y,4) = 3*stone) and empty(a-1,y)   and empty(a+4,y)  ) then n1 := 1;
    if (vert(x,b,5) = 4*stone) or((vert(x,b,4) = 3*stone) and empty(x,b-1)   and empty(x,b+4)  ) then n2 := 1;
    if (up(a,b,5)   = 4*stone) or((up  (a,b,4) = 3*stone) and empty(a-1,b-1) and empty(a+4,b+4)) then n3 := 1;
    if (down(a,c,5) = 4*stone) or((down(a,c,4) = 3*stone) and empty(a-1,c+1) and empty(a+4,c-4)) then n4 := 1;
  end;
  Result := n1+n2+n3+n4;
end;


{ ***********************************
  Dealing with backtracking
  *********************************** }

var xpos,ypos:array[1..FSIZE*FSIZE] of integer; // The list of still possible moves
    npos:integer;                               // Their number
     steps,maxsteps:integer;                    // Number of steps allowed in the recursion

function TrackBlack(d:integer):integer; forward;

function TrackWhite(d:integer):integer;
var x,y,i:integer;
begin
  Result := -1;
  steps := steps+1;
  if steps > maxsteps then Exit;
  for i:=1 to npos do
  begin
    x := xpos[i]; y := ypos[i];
    if field[x,y] <> POS_EMPTY then Continue;
    field[x,y] := POS_BLACK;               // check whether black can win here
    if CheckKin5(x,y,POS_BLACK,5) > 0 then // is so, this is a forced move for white
    begin                                  // on which we will recurse if it is not
      field[x,y] := POS_WHITE;             // a force for black itself
      if CheckKin5(x,y,POS_WHITE,4) = 0 then
        Result := 0
      else
        Result := TrackBlack(d);
      field[x,y] := POS_EMPTY;
      Exit;
    end;
    field[x,y] := POS_EMPTY;
  end;
  Result := 0;
end;

function TrackBlack(d:integer):integer;
var x,y,i,best:integer;
begin
  Result := -1;
  steps := steps+1;
  if steps > maxsteps then Exit;
  best := 0;
  // Check for clear winners and forced moves and recurse on them
  for i:=1 to npos do
  begin
    x := xpos[i]; y := ypos[i];
    if field[x,y] <> POS_EMPTY then Continue;
    field[x,y] := POS_BLACK;
    if CheckKin5(x,y,POS_BLACK,4) > 0 then    // this will create a force for white
    begin
      if CheckThrees(x,y,POS_BLACK) > 1 then  // it is a clear winner
        begin field[x,y] := POS_EMPTY; Result := 10+d; Exit; end;
      if (d > 0) then
      begin
        Result := TrackWhite(d-1);
        field[x,y] := POS_EMPTY;
        if Result = -1 then Exit;
        if Result > 10 then Exit;
        if Result > best then best := Result;
      end;
    end;
    field[x,y] := POS_EMPTY;
  end;
  // Check for almost certain winners but don't recurse on them
  Result := 0;
  for i:=1 to npos do
  begin
    x := xpos[i]; y := ypos[i];
    if field[x,y] <> POS_EMPTY then Continue;
    field[x,y] := POS_BLACK;
    if CheckThrees(x,y,POS_BLACK) > 1 then Result := d;
    field[x,y] := POS_EMPTY;
    if Result > 0 then Exit;
  end;
  Result := best;
end;

function PerformTrack(x,y,d:integer):integer;
// Does the backtracking with the given depth
begin
  Result := 0;
  field[x,y] := POS_BLACK;
  if CheckKin5(x,y,POS_BLACK,4) > 0 then
  begin
    steps := 0;
    maxsteps := 5000;
    Result := TrackWhite(d);
  end;
  field[x,y] := POS_EMPTY;
end;

procedure DoTrack(d:integer);
// Does the backtracking on all moved
var i,j,best,res:integer;
begin
  best := 0;
  j := 0;
  for i:=1 to npos do
  begin
    res := PerformTrack(xpos[i],ypos[i],d);
    if res > best then
      begin best := res; j := 0; end;
    if res = best then
      begin j := j+1; xpos[j] := xpos[i]; ypos[j] := ypos[i]; end;
  end;
  npos := j;
end;


{ ***********************************
  Computing moves
  *********************************** }

function PerformTest(n,x,y,stone:integer):integer;
// Does the test n for position x,y and the stone and returns the quality
var player,other:integer;
begin
  Result := 0;
  if stone = POS_WHITE then
    begin player := POS_WHITE; other := POS_BLACK; end
  else
    begin player := POS_BLACK; other := POS_WHITE; end;
  if odd(n) then
    field[x,y] := player
  else
    field[x,y] := other;
  case n of
  1: Result := checkKin5(x,y,player,5);
  2: Result := checkKin5(x,y,other,5);
  3: begin
       Result := checkKin5(x,y,player,4);
       if (Result = 1) and (CheckThrees(x,y,player) <= 1) then Result := 0;
     end;
  4: begin
       Result := checkKin5(x,y,other,4);
       if (Result = 1) and (CheckThrees(x,y,other) <= 1) then Result := 0;
     end;
  5: begin
       Result := CheckThrees(x,y,player);
       if (Result <= 1) or (movenumb < 6) then Result := 0;
     end;
  6: begin
       Result := CheckThrees(x,y,other);
       if (Result <= 1) or (movenumb < 5) then Result := 0;
     end;
  7: begin
       Result := CheckThrees(x,y,player);
       if (movenumb < 6) then Result := 0;
     end;
  8: begin
       Result := CheckThrees(x,y,other);
       if (movenumb < 5) then Result := 0;
     end;
  9: begin
       Result := CheckKin3(x,y,player,3);
       if (movenumb < 6) then Result := 0;
     end;
 10: begin
       Result := CheckKin3(x,y,other,3);
       if (movenumb < 5) then Result := 0;
     end;
 11: Result := checkKin4(x,y,player,2);
 12: Result := checkKin4(x,y,other,2);
  end;
  field[x,y] := POS_EMPTY;
end;

procedure DoTest(n,stone:integer);
// Do a particular test on all the possible moves
var i,j,best,res:integer;
begin
  best := 0;
  j := 0;
  for i:=1 to npos do
  begin
    res := PerformTest(n,xpos[i],ypos[i],stone);
    if res > best then
      begin best := res; j := 0; end;
    if res = best then
      begin j := j+1; xpos[j] := xpos[i]; ypos[j] := ypos[i]; end;
  end;
  npos := j;
end;

procedure FindMove(stone,level:integer; var x,y:integer);
// Finds the best move for the stone for the level and returns it in x,y
var pos:array[1..FSIZE,1..FSIZE] of boolean;
    i,j,k,l,n:integer;
begin
  // Find the reasonable possible moves
  for i:=1 to FSIZE do
    for j:=1 to FSIZE do
      pos[i,j] := false;
  for i:=1 to FSIZE do
    for j:=1 to FSIZE do
      if field[i,j] <> POS_EMPTY then
        for k:=i-2 to i+2 do
          for l:=j-2 to j+2 do
            if field[k,l] = POS_EMPTY then
              pos[k,l] := true;
  npos := 0;
  for i:=1 to FSIZE do
    for j:=1 to FSIZE do
      if pos[i,j] then
      begin
        npos := npos+1;
        xpos[npos] := i;
        ypos[npos] := j;
      end;
  // Now always do the first 3 tests
  for n:=1 to 3 do DoTest(n,stone);
  // Do backtracking if level is 3  and black is moving
  if (level = 3) and (stone = POS_BLACK) then
  begin
    DoTrack(4);
  end;
  // Now do the appropriate other tests
  if level = 1 then
    begin DoTest(4,stone); DoTest(5,stone); DoTest(6,stone); end
  else if level = 2 then
    for n:=4 to 9 do DoTest(n,stone)
  else
    for n:=4 to 10 do DoTest(n,stone);
  DoTest(11,stone); DoTest(12,stone);
  // Now pick a remaining move
  if npos = 0 then
    begin x := FSIZE div 2 + 1; y := FSIZE div 2 + 1; end
  else
    begin i := random(npos)+1; x := xpos[i]; y := ypos[i]; end;
end;


{ ***********************************
  Dealing with the field
  *********************************** }

procedure InitField();
// Initializes an empty field
var i,j:integer;
begin
  for i:=-4 to FSIZE + 5 do
    for j:=-4 to FSIZE + 5 do
      field[i,j] := POS_BORDER;
  for i:=1 to FSIZE do
    for j:=1 to FSIZE do
      field[i,j] := POS_EMPTY;
  movenumb := 0;
end;

function Place(x,y,stone:integer):boolean;
// Places the stone; only for black or white stones; returns whether successful
begin
  Result := false;
  if (x<1) or (x>FSIZE) or (y<1) or (y>FSIZE) then Exit;
  if field[x,y] <> POS_EMPTY then Exit;
  if (stone <> POS_BLACK) and (stone <> POS_WHITE) then Exit;
  field[x,y] := stone;
  movenumb := movenumb+1;
  xmoves[movenumb] := x; ymoves[movenumb] := y;
  Result := true;
end;

function UndoLast():boolean;
begin
  Result := false;
  if movenumb = 0 then Exit;
  field[xmoves[movenumb],ymoves[movenumb]] := POS_EMPTY;
  movenumb := movenumb -1;
end;


{ ***********************************
  The external functions
  *********************************** }

function GB_Init_Field():real; cdecl;
// Initializes the field; returns whether correctly
begin
  InitField();
  Result := 1;
end;

function GB_Place_Stone(player,x,y:real):real; cdecl;
// Places a stone of player (1,2) at position x,y; Return whether successfull
var xx,yy:integer;
begin
  Result := 0;
  xx := Round(x); yy := Round(y);
  if (Round(player)<>1) and (Round(player)<>2) then Exit;
  if Round(player) = 1 then
    Result := Ord(Place(xx,yy,POS_WHITE))
  else
    Result := Ord(Place(xx,yy,POS_BLACK));
end;

function GB_Undo_Last():real; cdecl;
// Undoes the last made move; Returns whether a move existed
begin
  Result := Ord(UndoLast());
end;

function GB_Get_Stone(x,y:real):real; cdecl;
// Returns the stone at position x,y (-1 = outside, 0=empty, 1,2=player)
var xx,yy:integer;
begin
 Result := -1;
 xx := Round(x); yy := Round(y);
 if (xx<1) or (xx>FSIZE) or (yy<1) or (yy>FSIZE) then Exit;
 if field[xx,yy] = POS_EMPTY then Result := 0;
 if field[xx,yy] = POS_WHITE then Result := 1;
 if field[xx,yy] = POS_BLACK then Result := 2;
end;

function GB_Check_Empty(x,y:real):real; cdecl;
// Returns whether position (x,y) is empty
begin
 Result := Ord(GB_Get_Stone(x,y) = 0);
end;

function GB_Check_Win(x,y:real):real; cdecl;
// Returns whether the last stone placed on position (x,y) leads to a win
var xx,yy:integer;
begin
  Result := 0;
  xx := Round(x); yy := Round(y);
 if (xx<1) or (xx>FSIZE) or (yy<1) or (yy>FSIZE) then Exit;
 if field[xx,yy] = POS_EMPTY then Exit;
 Result := Ord(CheckWin(xx,yy));
end;

function GB_Check_Draw():real; cdecl;
// Returns whether there is a draw
begin
  Result := Ord(movenumb = FSIZE*FSIZE);
end;

function GB_Get_Last():real; cdecl;
// Returns the last played stone position
begin
  if movenumb = 0 then
    Result := -1
  else
    Result := 100*xmoves[movenumb] + ymoves[movenumb];
end;

function GB_Get_Win_Pos(n:real):real; cdecl;
// Returns the n'th winning position in the form 100*x+y
begin
   Result := 100*winx[Round(n)]+winy[Round(n)];
end;

function GB_Find_Move(player,level:real):real; cdecl;
// Returns the best move for the player at the indicated playing level
// It does not make the move
var xx,yy:integer;
begin
  if Round(player) = 1 then
    FindMove(POS_WHITE,Round(level),xx,yy)
  else
    FindMove(POS_BLACK,Round(level),xx,yy);
  Result := 100*xx+yy;
end;

exports GB_Init_Field,
        GB_Place_Stone,
        GB_Undo_Last,
        GB_Get_Stone,
        GB_Check_Empty,
        GB_Check_Win,
        GB_Check_Draw,
        GB_Get_Last,
        GB_Get_Win_Pos,
        GB_Find_Move;

begin
end.
